package net.kldp.OpenSigner.utils;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.KeyStore;
import java.security.MessageDigest;
import java.security.PrivateKey;
import java.security.Signature;

import net.kldp.OpenSigner.applet.Signer;

public class xmldsigUtil {
    public static final String NS_OPENSIGNER_SIGNATURE = "http://open-signer.kldp.net/2007/04/signature#";
    public static final String NS_W3_XMLDSIG = "http://www.w3.org/2000/09/xmldsig#";
    static final int BASE64_LINELENGTH = 76;
    

    /**
     * Get xmldsig signature message from parameters
     * @param keyStore 
     *
     * @param visibleProperties
     * @param invisibleProperties
     * @param attachments
     * @return StringBuffer
     * @throws java.security.NoSuchAlgorithmException
     *
     * @throws java.io.UnsupportedEncodingException
     *
     * @throws java.security.NoSuchProviderException
     *
     */
public static StringBuffer getSignature(PrivateKey user_privatekey, 
    		KeyStore keyStore, char[] password, 
    		StringBuffer signText) 
    	throws IOException, Exception {
	
        StringBuffer sb = new StringBuffer();
        sb.append("<?xml version=\"1.0\" encoding=\"utf8\" ?>\n");
        sb.append("<opensigner:signature xmlns:opensigner=\""+NS_OPENSIGNER_SIGNATURE+"\" version=\"0.1\">");
        sb.append("<ds:Signature xmlns:ds=\""+NS_W3_XMLDSIG+"\">\n");

        StringBuffer objectText = getObject(signText);

        StringBuffer signedInfo = getSignedInfo(objectText);

        sb.append(signedInfo.toString()).append("\n");
        sb.append(getSignatureValue(user_privatekey, password, signedInfo).toString());
        String keyInfo = getKeyInfo(keyStore, password);
        sb.append(keyInfo);

        sb.append(objectText.toString()).append("\n");

        sb.append("</ds:Signature>");
        sb.append("</opensigner:signature>");
        return sb;
    }

/**
 * Get the Object section of the xmldsig message. Properties are encoded as NAME VALUE
 * pairs, where name and value are encoded ad BASE64, to avoid XML message syntaxt conflicts.
 * When decoded, signtext section sould be likewise decoded.
 * The invisible properties that have not been visible to the signer, whereas the visible
 * properties have been visible.
 *
 * @param visibleProperties
 * @param signText            the tbs text
 * @param invisibleProperties
 * @return StringBuffer
 */
private static StringBuffer getObject(StringBuffer signText) throws UnsupportedEncodingException {
    StringBuffer sb = new StringBuffer();
    sb.append("<ds:Object xmlns:ds=\""+NS_W3_XMLDSIG+"\" xmlns:opensigner=\""+NS_OPENSIGNER_SIGNATURE+"\" Id=\"ToBeSigned\">");
    sb.append("<ds:SignatureProperties>\n");
    sb.append(getPropertiesFragment(signText, true).toString());
    sb.append("</ds:SignatureProperties>");
    sb.append("</ds:Object>");
    return sb;
}

/**
 * Get SignedInfo section of xmldsig
 *
 * @param toBeSigned
 * @param attachmentObjects
 * @return StringBuffer
 * @throws java.io.UnsupportedEncodingException
 *
 */
private static StringBuffer getSignedInfo(StringBuffer toBeSigned) throws Exception {
    StringBuffer sb = new StringBuffer();
    sb.append("<ds:SignedInfo xmlns:ds=\""+NS_W3_XMLDSIG+"\" xmlns:opensigner=\""+NS_OPENSIGNER_SIGNATURE+"\">\n");
    sb.append("<ds:CanonicalizationMethod Algorithm=\"http://www.w3.org/TR/2001/REC-xml-c14n-20010315\"></ds:CanonicalizationMethod>\n");
    sb.append("<ds:SignatureMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#rsa-sha1\"></ds:SignatureMethod>\n");

    sb.append("<ds:Reference URI=\"#ToBeSigned\">\n");
    sb.append("<ds:DigestMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#sha1\"></ds:DigestMethod>\n");
    sb.append("<ds:DigestValue>");
    sb.append(getDigestValue(toBeSigned));
    sb.append("</ds:DigestValue>\n");
    sb.append("</ds:Reference>\n");

    sb.append("</ds:SignedInfo>");
    return sb;
}

/**
 * Get SignatureValue section of xmldsig
 *
 * @param signedInfo
 * @return String
 * @throws java.security.InvalidKeyException
 *
 * @throws java.security.SignatureException
 *
 * @throws java.security.NoSuchAlgorithmException
 *
 * @throws UnsupportedEncodingException
 * @throws java.security.NoSuchProviderException
 *
 */
private static StringBuffer getSignatureValue(PrivateKey pk, char[] password, StringBuffer signedInfo) throws IOException, Exception {
    StringBuffer sb = new StringBuffer();
    byte[] data = signedInfo.toString().getBytes("UTF8");

    Signature sig = Signature.getInstance("SHA1withRSA");
    sig.initSign(pk);
    sig.update(data);
    byte[] bSignedValue = sig.sign();
    
    /*
    SignUtil su = new SignUtil();
    su.signInit(pk);
    su.signUpdate(data);
    byte[] bSignedValue = su.signFinal();
    */
    
    sb.append("<ds:SignatureValue>\n");
    // sb.append(wrapLines(new String(Base64.encode(JCETools.sign(Signature.getInstance("SHA1withRSA"), data, pk))), BASE64_LINELENGTH).toString());
    sb.append(StringUtil.wrapLines(new String(Base64.encode(bSignedValue)), BASE64_LINELENGTH).toString());
    sb.append("\n</ds:SignatureValue>\n");
    return sb;
}

/**
 * Get KEYINFO section of xmldsig
 *
 * @return String
 */
private static String getKeyInfo(KeyStore keyStore, char[] password) throws IOException, Exception {
    StringBuffer sb = new StringBuffer();
    sb.append("<ds:KeyInfo>\n");
    sb.append(getX509Data(keyStore));
    sb.append("</ds:KeyInfo>\n");
    return sb.toString();
}

/**
 * Get X509Data section of xmldsig
 *
 * @return String
 */
private static String getX509Data(KeyStore keyStore) throws IOException, Exception {
    StringBuffer sb = new StringBuffer();
    byte[][] chain = new byte[keyStore.getCertificateChain(Signer.alias_in_pkcs12).length][0];
    for (int i = 0; i < chain.length; i++) {
    	chain[i] = keyStore.getCertificateChain(Signer.alias_in_pkcs12)[i].getEncoded();
    }
    
    if (chain != null) {
        for (int i = 0; i < chain.length; i++) {
            sb.append("<ds:X509Data>\n");
            sb.append(getX509Certificate(chain[i]));
            sb.append("</ds:X509Data>\n");
        }
    }

    return sb.toString();
}

/**
 * Get X509Data section of xmldsig
 *
 * @param cert
 * @return String
 */
private static String getX509Certificate(byte[] cert) {
    StringBuffer sb = new StringBuffer();
    sb.append("<ds:X509Certificate>\n");
    sb.append(StringUtil.wrapLines(new String(Base64.encode(cert)), BASE64_LINELENGTH).toString());
    sb.append("\n</ds:X509Certificate>\n");
    return sb.toString();
}
private static StringBuffer getPropertiesFragment(StringBuffer signText, boolean visible) throws UnsupportedEncodingException {
    StringBuffer s = new StringBuffer();
    if (signText != null) {
        s.append(getSignatureProperty("signtext", signText, true).toString()).append("\n");
    }

    return s;
}

/**
 * Generate a <SignatureProperty> node, encode name and value in order to support special chars
 * s
 *
 * @param name
 * @param value
 * @param setToBeSignedId
 * @return the <SignatureProperty> node
 */

private static StringBuffer getSignatureProperty(String name, StringBuffer value, boolean setToBeSignedId) throws UnsupportedEncodingException {
    StringBuffer sb = new StringBuffer();

/*
        Better safe than sorry:
        An extra check in case we are not escaping all necessary characters correctly
    */

    StringBuffer s;
    if (value == null || value.length() == 0) {
        s = new StringBuffer("");
    } else {
        s = new StringBuffer(new String(Base64.encode(value.toString().getBytes("UTF8"))));
    }

    sb.append("<ds:SignatureProperty>");
    sb.append("<opensigner:Name>");
    sb.append(name);
    sb.append("</opensigner:Name>");
    sb.append("<opensigner:Value Encoding=\"base64\" VisibleToSigner=\"").append(setToBeSignedId ? "yes" : "no").append("\">");
    sb.append(s.toString());
    sb.append("</opensigner:Value>");
    sb.append("</ds:SignatureProperty>");
    return sb;
}

/**
 * Get Digest from signtext, coded as BASE64
 *
 * @param signText
 * @return String
 * @throws java.security.NoSuchProviderException
 *
 * @throws java.security.NoSuchAlgorithmException
 *
 * @throws UnsupportedEncodingException
 */
private static String getDigestValue(StringBuffer signText) throws Exception {
    byte[] data = signText.toString().getBytes("UTF8");
    MessageDigest md = MessageDigest.getInstance("SHA1");
    md.reset();
    byte[] hash = md.digest(data);
    return new String(Base64.encode(hash));
}

}
